@@ -2,6 +2,7 @@ import collections
 import six
 
 from . import logical
+from .dict_wrapper import DictWrapper
 from avro import schema
 from avro import io
 
@@ -44,11 +45,45 @@ class AvroJsonConverter(object):
                     False not in
                     [self.validate(expected_schema.values, v, skip_logical_types) for v in datum.values()])
         elif schema_type in ['union', 'error_union']:
+            if isinstance(datum, DictWrapper):
+                # Match the type based on the declared schema.
+                data_schema = self._get_record_schema_if_available(datum)
+                for i, candidate_schema in enumerate(expected_schema.schemas):
+                    if candidate_schema.namespace == data_schema.namespace \
+                        and candidate_schema.name == data_schema.name:
+                        return self.validate(candidate_schema, datum)
+
+            # If the union type is using a "name" to distinguish the type, we
+            # must handle this specially during validation.
+            value_type = None
+            if not self.fastavro and isinstance(datum, dict):
+                if len(datum) == 1:
+                    items = list(six.iteritems(datum))
+                    if not items:
+                        return None
+                    value_type = items[0][0]
+                    value = items[0][1]
+            elif self.fastavro and (isinstance(datum, list) or isinstance(datum, tuple)):
+                if len(datum) == 2:
+                    value_type = datum[0]
+                    value = datum[1]
+            if value_type is not None:
+                for s in expected_schema.schemas:
+                    name = self._fullname(s)
+                    if name == value_type:
+                        if self.validate(s, value, skip_logical_types):
+                            return True
+                        # If the specialized validation fails, we still attempt normal validation.
+
             return True in [self.validate(s, datum, skip_logical_types) for s in expected_schema.schemas]
         elif schema_type in ['record', 'error', 'request']:
-            return (isinstance(datum, dict) and
+            return ((isinstance(datum, dict) or isinstance(datum, DictWrapper)) and
                     False not in
                     [self.validate(f.type, datum.get(f.name), skip_logical_types) for f in expected_schema.fields])
+        elif not self.fastavro and schema_type == 'bytes':
+            # Specialization for bytes, which are encoded as strings in JSON.
+            if isinstance(datum, str):
+                return True
 
         return io_validate(expected_schema, datum)
 
@@ -89,7 +124,7 @@ class AvroJsonConverter(object):
             return type(data_obj).RECORD_SCHEMA
         return None
 
-    def _generic_to_json(self, data_obj, writers_schema):
+    def _generic_to_json(self, data_obj, writers_schema, was_within_array=False):
         if self.use_logical_types and writers_schema.props.get('logicalType'):
             lt = self.logical_types.get(writers_schema.props.get('logicalType'))  # type: logical.LogicalTypeProcessor
             if lt.can_convert(writers_schema):
@@ -112,13 +147,15 @@ class AvroJsonConverter(object):
         elif writers_schema.type in ['record', 'error', 'request']:
             result = self._record_to_json(data_obj, writers_schema)
         elif writers_schema.type in ['union', 'error_union']:
-            result = self._union_to_json(data_obj, writers_schema)
+            result = self._union_to_json(data_obj, writers_schema, was_within_array)
         else:
             raise schema.AvroException('Invalid schema type: %s' % writers_schema.type)
 
         return result
 
     def _primitive_to_json(self, data_obj, writers_schema):
+        if not self.fastavro and isinstance(data_obj, bytes):
+            return data_obj.decode()
         return data_obj
 
     def _fixed_to_json(self, data_obj, writers_schema):
@@ -128,7 +165,7 @@ class AvroJsonConverter(object):
         return data_obj
 
     def _array_to_json(self, data_obj, writers_schema):
-        return [self._generic_to_json(x, writers_schema.items) for x in data_obj]
+        return [self._generic_to_json(x, writers_schema.items, True) for x in data_obj]
 
     def _map_to_json(self, data_obj, writers_schema):
         return {name: self._generic_to_json(x, writers_schema.values) for name, x in six.iteritems(data_obj)}
@@ -144,15 +181,23 @@ class AvroJsonConverter(object):
         return result
     
     def _is_unambiguous_union(self, writers_schema) -> bool:
+        if any(isinstance(candidate_schema, schema.EnumSchema) for candidate_schema in writers_schema.schemas):
+            if len(writers_schema.schemas) == 2 and any(candidate_schema.type == 'null' for candidate_schema in writers_schema.schemas):
+                # Enums and null do not conflict, so this is fine.
+                return True
+            else:
+                # Enum and string conflict, so this case is ambiguous.
+                return False
+
         advanced_count = 0
         for candidate_schema in writers_schema.schemas:
-            if not isinstance(candidate_schema, schema.PrimitiveSchema):
+            if candidate_schema.type != 'null':
                 advanced_count += 1
         if advanced_count <= 1:
             return True
         return False
 
-    def _union_to_json(self, data_obj, writers_schema):
+    def _union_to_json(self, data_obj, writers_schema, was_within_array=False):
         index_of_schema = -1
         data_schema = self._get_record_schema_if_available(data_obj)
         for i, candidate_schema in enumerate(writers_schema.schemas):
@@ -174,9 +219,10 @@ class AvroJsonConverter(object):
             return None
         
         output_obj = self._generic_to_json(data_obj, candidate_schema)
-        if self._is_unambiguous_union(writers_schema):
+        if not self.fastavro and not was_within_array and self._is_unambiguous_union(writers_schema):
             # If the union is unambiguous, we can avoid wrapping it in
-            # an extra layer of tuples or dicts.
+            # an extra layer of tuples or dicts. Fastavro doesn't like this though.
+            # Arrays with unions inside must specify the type.
             return output_obj
         if self.fastavro:
             # Fastavro likes tuples instead of dicts for union types.
@@ -220,6 +266,9 @@ class AvroJsonConverter(object):
         return result
 
     def _primitive_from_json(self, json_obj, writers_schema, readers_schema):
+        if not self.fastavro and writers_schema.type == 'bytes':
+            if isinstance(json_obj, str):
+                return json_obj.encode()
         return json_obj
 
     def _fixed_from_json(self, json_obj, writers_schema, readers_schema):
@@ -241,7 +290,7 @@ class AvroJsonConverter(object):
             return None
         value_type = None
         value = None
-        if not self.fastavro and isinstance(json_obj, collections.Mapping):
+        if not self.fastavro and isinstance(json_obj, dict):
             items = list(six.iteritems(json_obj))
             if not items:
                 return None
@@ -262,16 +311,21 @@ class AvroJsonConverter(object):
             if self.validate(s, json_obj, skip_logical_types=True):
                 return self._generic_from_json(json_obj, s, readers_schema)
         raise schema.AvroException('Datum union type not in schema: %s', value_type)
+    
+    def _make_type(self, tp, record):
+        if issubclass(tp, DictWrapper):
+            return tp.construct(record)
+        return tp(record)
 
     def _instantiate_record(self, decoded_record, writers_schema, readers_schema):
         # First try the fullname, which includes namespaces.
         readers_name = self._fullname(readers_schema)
         if readers_name in self.schema_types:
-            return self.schema_types[readers_name](decoded_record)
+            return self._make_type(self.schema_types[readers_name], decoded_record)
         # Fallback to the bare name, without namespace.
         readers_name = readers_schema.name
         if readers_name in self.schema_types:
-            return self.schema_types[readers_name](decoded_record)
+            return self._make_type(self.schema_types[readers_name], decoded_record)
         return decoded_record
 
     def _record_from_json(self, json_obj, writers_schema, readers_schema):
