@@ -1,3 +1,4 @@
+from typing import Dict
 import os
 
 from avro import schema
@@ -37,10 +38,10 @@ def clean_fullname(fullname):
     return fullname
 
 
-def convert_default(full_name, idx, do_json=True):
+def convert_default(idx, full_name=None, do_json=True):
     if do_json:
-        return (f'_json_converter.from_json_object({full_name}Class.RECORD_SCHEMA.fields[{idx}].default,'
-               + f' writers_schema={full_name}Class.RECORD_SCHEMA.fields[{idx}].type)')
+        return (f'_json_converter.from_json_object(self.RECORD_SCHEMA.field_map["{idx}"].default,'
+               + f' writers_schema=self.RECORD_SCHEMA.field_map["{idx}"].type)')
     else:
         return f'self.RECORD_SCHEMA.field_map["{idx}"].default'
 
@@ -62,8 +63,8 @@ def get_default(field, use_logical_types, my_full_name=None, f_name=None):
                                                 schema.RecordSchema)))
             return v
         elif isinstance(default_type, schema.RecordSchema):
-            d = convert_default(idx=i, full_name=my_full_name, do_json=True)
-            return f'{field.name}Class({d})'
+            d = convert_default(idx=f_name, do_json=True)
+            return d
         elif isinstance(default_type, (schema.PrimitiveSchema, schema.EnumSchema, schema.FixedSchema)):
             d = convert_default(full_name=my_full_name, idx=f_name, do_json=False)
             return d
@@ -88,10 +89,10 @@ def get_default(field, use_logical_types, my_full_name=None, f_name=None):
         elif isinstance(default_type, schema.ArraySchema):
             return 'list()'
         elif isinstance(default_type, schema.FixedSchema):
-            return 'str()'
+            return 'bytes()'
         elif isinstance(default_type, schema.RecordSchema):
             f = clean_fullname(default_type.name)
-            return f'{f}Class()'
+            return f'{f}Class.construct_with_defaults()'
     raise AttributeError('cannot get default for field')
 
 def write_defaults(record, writer, my_full_name=None, use_logical_types=False):
@@ -152,9 +153,8 @@ def {name}(self) -> {ret_type_name}:
     {get_docstring}
     return self._inner_dict.get('{raw_name}')  # type: ignore
 
-
 @{name}.setter
-def {name}(self, value: {ret_type_name}):
+def {name}(self, value: {ret_type_name}) -> None:
     {set_docstring}
     self._inner_dict['{raw_name}'] = value
 
@@ -281,7 +281,7 @@ def write_preamble(writer, use_logical_types, custom_imports):
         writer.write('from avrogen import logical\n')
     writer.write('from avro.schema import RecordSchema, SchemaFromJSONData as make_avsc_object\n')
     writer.write('from avro import schema as avro_schema\n')
-    writer.write('from typing import List, Dict, Union, Optional, overload\n')
+    writer.write('from typing import List, Dict, Union, Optional\n')
     writer.write('\n')
 
 
@@ -341,7 +341,12 @@ def write_reader_impl(record_types, writer, use_logical_types):
                 '\nresult = super(SpecificDatumReader, self).read_record(writers_schema, readers_schema, decoder)')
             writer.write('\n\nif readers_schema.fullname in SpecificDatumReader.SCHEMA_TYPES:')
             with writer.indent():
-                writer.write('\nresult = SpecificDatumReader.SCHEMA_TYPES[readers_schema.fullname](result)')
+                writer.write('\ntp = SpecificDatumReader.SCHEMA_TYPES[readers_schema.fullname]')
+                writer.write('\nif issubclass(tp, DictWrapper):')
+                writer.write('\n    result = tp.construct(result)')
+                writer.write('\nelse:')
+                writer.write('\n    # tp is an enum')
+                writer.write('\n    result = tp(result)  # type: ignore')
             writer.write('\n\nreturn result')
 
 
@@ -393,21 +398,25 @@ def write_schema_record(record, writer, use_logical_types):
 
         write_record_init(record, writer, use_logical_types)
 
-        write_serialization_stubs(record, writer, use_logical_types)
-
         write_fields(record, writer, use_logical_types)
 
 
 def write_record_init(record, writer, use_logical_types):
-    writer.write('\n\n@overload')
     writer.write('\ndef __init__(self,')
     with writer.indent():
         delayed_lines = []
+        default_map: Dict[str, str] = {}
         for field in record.fields:  # type: schema.Field
             name = get_field_name(field, use_logical_types)
             ret_type_name = get_field_type_name(field.type, use_logical_types)
             default_type, nullable = find_type_of_default(field.type)
 
+            if not nullable and field.has_default:
+                # print(record.name, field.name, field.default)
+                default = get_default(field, use_logical_types, f_name=field.name)
+                default_map[name] = default
+                ret_type_name = f"Optional[{ret_type_name}]"
+                nullable = True
             if nullable:
                 delayed_lines.append(f'\n{name}: {ret_type_name}=None,')
             else:
@@ -417,46 +426,35 @@ def write_record_init(record, writer, use_logical_types):
         for line in delayed_lines:
             writer.write(line)
     writer.write('\n):')
-    with writer.indent():
-        writer.write('\n...')
-
-    writer.write('\n\n@overload')
-    writer.write('\ndef __init__(self, _inner_dict: Optional[dict]=None):')
-    with writer.indent():
-        writer.write('\n...')
-
-    writer.write('\n\ndef __init__(self, _inner_dict=None, **kwargs):')
     with writer.indent():
         writer.write('\n')
-        writer.write('super({name}Class, self).__init__({{}})'.format(name=record.name))
+        writer.write('super().__init__()')
+        writer.write('\n')
 
+        for field in record.fields:  # type: schema.Field
+            name = get_field_name(field, use_logical_types)
+            if name in default_map:
+                writer.write(f'\nif {name} is None:')
+                writer.write(f'\n    # default: {repr(field.default)}')
+                writer.write(f'\n    self.{name} = {default_map[name]}')
+                writer.write(f'\nelse:')
+                writer.write(f'\n    self.{name} = {name}')
+            else:
+                writer.write(f'\nself.{name} = {name}')
+
+    writer.write('\n\n@classmethod')
+    writer.write(f'\ndef construct_with_defaults(cls) -> "{record.name}Class":')
+    with writer.indent():
+        writer.write('\nself = cls.construct({})')
+        writer.write('\nself._restore_defaults()')
+        writer.write('\n')
+        writer.write('\nreturn self')
+
+    writer.write('\n')
+    writer.write(f'\ndef _restore_defaults(self) -> None:')
+    with writer.indent():
         write_defaults(record, writer, use_logical_types=use_logical_types)
 
-        writer.write('\nif _inner_dict is not None:')
-        with writer.indent():
-            writer.write('\nfor key, value in _inner_dict.items():')
-            with writer.indent():
-                writer.write('\ngetattr(self, key)')
-                writer.write('\nsetattr(self, key, value)')
-        writer.write('\nfor key, value in kwargs.items():')
-        with writer.indent():
-            writer.write('\nif value is not None:')
-            with writer.indent():
-                writer.write('\ngetattr(self, key)')
-                writer.write('\nsetattr(self, key, value)')
-
-
-def write_serialization_stubs(record, writer, use_logical_types):
-    # writer.write(f'\n\n@classmethod')
-    # writer.write(f'\ndef from_obj(cls, obj: dict, tuples=False) -> "{record.name}Class":')
-    # with writer.indent():
-    #     writer.write('\n...')
-
-    # writer.write(f'\n\ndef to_obj(self, tuples=False) -> dict:')
-    # with writer.indent():
-    #     writer.write('\n...')
-    pass
-
 
 def write_enum(enum, writer):
     """
@@ -476,6 +474,11 @@ def write_enum(enum, writer):
             writer.write('# No docs available.')
 
         writer.write('\n\n')
+        symbolDocs = enum.other_props.get('symbolDocs', {})
         for field in enum.symbols:
+            if symbolDocs:
+                writer.write('\n')
+            if field in symbolDocs:
+                writer.write(f'"""{symbolDocs[field]}"""\n')
             writer.write('{name} = "{name}"\n'.format(name=field))
         writer.write('\n')
